@@ -1,12 +1,16 @@
-# Copyright (C) 2024 qBraid
+# Copyright 2025 qBraid
 #
-# This file is part of pyqasm
+# Licensed under the Apache License, Version 2.0 (the "License");
+# you may not use this file except in compliance with the License.
+# You may obtain a copy of the License at
 #
-# Pyqasm is free software released under the GNU General Public License v3
-# or later. You can redistribute and/or modify it under the terms of the GPL v3.
-# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
+#     http://www.apache.org/licenses/LICENSE-2.0
 #
-# THERE IS NO WARRANTY for pyqasm, as per Section 15 of the GPL v3.
+# Unless required by applicable law or agreed to in writing, software
+# distributed under the License is distributed on an "AS IS" BASIS,
+# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
+# See the License for the specific language governing permissions and
+# limitations under the License.
 
 """
 Top level module containing the main PyQASM functionality.
@@ -23,6 +27,7 @@ Functions
    loads
    dump
    dumps
+   draw
 
 Classes
 ---------
@@ -46,17 +51,19 @@ Exceptions
 
 """
 import warnings
+from importlib.metadata import version
 
 try:
     # Injected in _version.py during the build process.
     from ._version import __version__  # type: ignore
-except ImportError:  # pragma: no cover
-    warnings.warn("Importing 'pyqasm' outside a proper installation.")
-    __version__ = "dev"
+except (ImportError, ModuleNotFoundError):  # pragma: no cover
+    warnings.warn("Importing 'pyqasm' outside a proper installation.", UserWarning)
+    __version__ = version("pyqasm")
 
 from .entrypoint import dump, dumps, load, loads
 from .exceptions import PyQasmError, QasmParsingError, ValidationError
 from .modules import Qasm2Module, Qasm3Module, QasmModule
+from .printer import draw
 
 __all__ = [
     "PyQasmError",
@@ -66,6 +73,7 @@ __all__ = [
     "loads",
     "dump",
     "dumps",
+    "draw",
     "QasmModule",
     "Qasm2Module",
     "Qasm3Module",
